#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
# !/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains the function which defines the images margins.
"""

from typing import Dict

import numpy as np
import xarray as xr


def get_margins(disp_min: int, disp_max: int, cfg: Dict[str, dict]) -> xr.DataArray:
    """
    Calculates the margins for the left and right images according to the configuration

    :param disp_min: minimal disparity
    :type disp_min: int
    :param disp_max: maximal disparity
    :type disp_max: int
    :param cfg: user configuration
    :type cfg: dict of dict
    :return: margin for the images, 2D (image, corner) DataArray, with the dimensions image = \
     ['left_margin', 'right_margin'], corner = ['left', 'up', 'right', 'down']
    :rtype: DataArray
    """
    margin = xr.DataArray(np.zeros((2, 4), dtype=int),
                          coords=[['left_margin', 'right_margin'], ['left', 'up', 'right', 'down']],
                          dims=['image', 'corner'])
    margin.name = 'Margins'

    # Margins for the left image and for the right image

    # Pandora margins depends on the steps configured
    if cfg['optimization']['optimization_method'] == 'sgm':
        # SGM margin includes the census, vfit and median filter margins
        sgm_margins = 40
        r_marg = [sgm_margins + disp_max, sgm_margins, sgm_margins - disp_min, sgm_margins]
        s_marg = [sgm_margins - disp_min, sgm_margins, sgm_margins + disp_max, sgm_margins]

    else:
        r_marg = np.array([disp_max, 0, -disp_min, 0])
        s_marg = np.array([-disp_min, 0, + disp_max, 0])

        if cfg['matching_cost']['window_size'] != 1:
            r_marg += int(cfg['matching_cost']['window_size'] / 2)
            s_marg += int(cfg['matching_cost']['window_size'] / 2)

        if cfg['refinement']['refinement_method'] == 'vfit':
            r_marg[0] += 1
            r_marg[2] += 1
            s_marg[0] += 1
            s_marg[2] += 1

        if cfg['filter']['filter_method'] == 'median':
            r_marg += int(cfg['filter']['filter_size'] / 2)
            s_marg += int(cfg['filter']['filter_size'] / 2)

    # Same margin for left and right: take the larger
    same_margin = list(map(lambda input: max(input[0], input[1]), zip(r_marg, s_marg)))
    margin.loc[dict(image='left_margin')] = same_margin
    margin.loc[dict(image='right_margin')] = same_margin

    # Save disp_min and disp_max
    margin.attrs['disp_min'] = disp_min
    margin.attrs['disp_max'] = disp_max

    return margin



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains all the parameters related to the validity mask, defining each bit.
"""

# INVALID POINTS CONSTANTS
# INVALID
PANDORA_MSK_PIXEL_INVALID = 0b01111000011
# if bit 0 activated : The pixel is invalid : border of left image, OR nodata in left image
PANDORA_MSK_PIXEL_LEFT_NODATA_OR_BORDER = (1 << 0)
# if bit 1 activated : The pixel is invalid : disparity range to explore is missing in right image, OR
#               nodata in right image
PANDORA_MSK_PIXEL_RIGHT_NODATA_OR_DISPARITY_RANGE_MISSING = (1 << 1)
# if bit 2 activated : Information : disparity range to explore is incomplete (borders reached in right image)
PANDORA_MSK_PIXEL_RIGHT_INCOMPLETE_DISPARITY_RANGE = (1 << 2)
# if bit 3 activated : Information : Computation stopped during pixelic step, under pixelic interpolation never ended
PANDORA_MSK_PIXEL_STOPPED_INTERPOLATION = (1 << 3)
# if bit 4 activated : Information : Occlusion was filled
PANDORA_MSK_PIXEL_FILLED_OCCLUSION = (1 << 4)
# if bit 5 activated : Information : Mismatch was filled
PANDORA_MSK_PIXEL_FILLED_MISMATCH = (1 << 5)
# if bit 6 activated : The pixel is invalid : invalidated by validity mask of left image
PANDORA_MSK_PIXEL_IN_VALIDITY_MASK_LEFT = (1 << 6)
# if bit 7 activated : The pixel is invalid : invalidated by validity mask of right image
PANDORA_MSK_PIXEL_IN_VALIDITY_MASK_RIGHT = (1 << 7)
# if bit 8 activated : The pixel is invalid : pixel in occlusion area
PANDORA_MSK_PIXEL_OCCLUSION = (1 << 8)
# if bit 9 activate : The pixel is invalid : mismatch
PANDORA_MSK_PIXEL_MISMATCH = (1 << 9)
# if bit 10 activate : Information : Nodata was filled
PANDORA_MSK_PIXEL_FILLED_NODATA = (1 << 10)



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains classes and functions associated to the subpixel refinement step.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Tuple, Callable

import numpy as np
import xarray as xr
from numba import njit, prange

import pandora.constants as cst


class AbstractRefinement():
    """
    Abstract Refinement class
    """
    __metaclass__ = ABCMeta

    subpixel_methods_avail = {}

    def __new__(cls, **cfg: dict) -> object:
        """
        Return the plugin associated with the refinement_method given in the configuration

        :param cfg: configuration {'refinement_method': value}
        :type cfg: dictionary
        """
        if cls is AbstractRefinement:
            if isinstance(cfg['refinement_method'], str):
                try:
                    return super(AbstractRefinement, cls).__new__(cls.subpixel_methods_avail[cfg['refinement_method']])
                except KeyError:
                    logging.error('No subpixel method named % supported', cfg['refinement_method'])
                    raise KeyError
            else:
                if isinstance(cfg['refinement_method'], unicode): # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractRefinement, cls).__new__(
                            cls.subpixel_methods_avail[cfg['refinement_method'].encode('utf-8')])
                    except KeyError:
                        logging.error('No subpixel method named % supported', cfg['refinement_method'])
                        raise KeyError
        else:
            return super(AbstractRefinement, cls).__new__(cls)

    def subpixel_refinement(self, cv: xr.Dataset, disp: xr.Dataset) -> None:
        """
        Subpixel refinement of disparities and costs.

        :param cv: the cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param disp: Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type disp: xarray.Dataset
        :return: None
        """
        d_min = cv.coords['disp'].data[0]
        d_max = cv.coords['disp'].data[-1]
        subpixel = cv.attrs['subpixel']
        measure = cv.attrs['type_measure']

        # Conversion to numpy array ( .data ), because Numba does not support Xarray
        itp_coeff, disp['disparity_map'].data, disp['validity_mask'].data = \
            self.loop_refinement(cv['cost_volume'].data, disp['disparity_map'].data, disp['validity_mask'].data,
                                 d_min, d_max, subpixel, measure, self.refinement_method)

        disp.attrs['refinement'] = self._refinement_method_name
        disp['interpolated_coeff'] = xr.DataArray(itp_coeff,
                                                  coords=[disp.coords['row'], disp.coords['col']],
                                                  dims=['row', 'col'])

    def approximate_subpixel_refinement(self, cv_left: xr.Dataset, disp_right: xr.Dataset) -> xr.Dataset:
        """
        Subpixel refinement of the right disparities map, which was created with the approximate method : a diagonal
        search for the minimum on the left cost volume

        :param cv_left: the left cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv_left: xarray.Dataset
        :param disp_right: right disparity map with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type disp_right: xarray.Dataset
        :return: disp_right Dataset with the variables :

                - disparity_map 2D xarray.DataArray (row, col) that contains the refined disparities
                - confidence_measure 3D xarray.DataArray (row, col, indicator) (unchanged)
                - validity_mask 2D xarray.DataArray (row, col) with the value of bit 3 ( Information: \
                calculations stopped at the pixel step, sub-pixel interpolation did not succeed )
                - interpolated_coeff 2D xarray.DataArray (row, col) that contains the refined cost
        :rtype: xarray.Dataset
        """
        d_min = cv_left.coords['disp'].data[0]
        d_max = cv_left.coords['disp'].data[-1]
        subpixel = cv_left.attrs['subpixel']
        measure = cv_left.attrs['type_measure']

        # Conversion to numpy array ( .data ), because Numba does not support Xarray
        itp_coeff, disp_right['disparity_map'].data, disp_right[
            'validity_mask'].data = self.loop_approximate_refinement(
            cv_left['cost_volume'].data, disp_right['disparity_map'].data, disp_right['validity_mask'].data, d_min,
            d_max,
            subpixel, measure, self.refinement_method)

        disp_right.attrs['refinement'] = self._refinement_method_name
        disp_right['interpolated_coeff'] = xr.DataArray(itp_coeff,
                                                        coords=[disp_right.coords['row'], disp_right.coords['col']],
                                                        dims=['row', 'col'])
        return disp_right

    @classmethod
    def register_subclass(cls, short_name: str):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.subpixel_methods_avail[short_name] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self) -> None:
        """
        Describes the subpixel method
        :return: None
        """
        print('Subpixel method description')

    @staticmethod
    @njit(parallel=True)
    def loop_refinement(cv: np.ndarray, disp: np.ndarray, mask: np.ndarray, d_min: int, d_max: int, subpixel: int,
                        measure: str, method: Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray, str],
                                                       Tuple[int, int, int]]) -> Tuple[np.ndarray, np.ndarray,
                                                                                       np.ndarray]:
        """
         Apply for each pixels the refinement method

        :param cv: cost volume to refine
        :type cv: 3D numpy array (row, col, disp)
        :param disp: disparity map
        :type disp: 2D numpy array (row, col)
        :param mask: validity mask
        :type mask: 2D numpy array (row, col)
        :param d_min: minimal disparity
        :type d_min: int
        :param d_max: maximal disparity
        :type d_max: int
        :param subpixel: subpixel precision used to create the cost volume
        :type subpixel: int ( 1 | 2 | 4 )
        :param measure: the measure used to create the cot volume
        :param measure: string
        :param method: the refinement method
        :param method: function
        :return: the refine coefficient, the refine disparity map, and the validity mask
        :rtype: tuple(2D numpy array (row, col), 2D numpy array (row, col), 2D numpy array (row, col))
         """
        n_row, n_col, _ = cv.shape
        itp_coeff = np.zeros((n_row, n_col), dtype=np.float64)

        for row in prange(n_row):
            for col in prange(n_col):
                # No interpolation on invalid points
                if (mask[row, col] & cst.PANDORA_MSK_PIXEL_INVALID) != 0:
                    itp_coeff[row, col] = np.nan
                else:
                    # conversion to numpy indexing
                    dsp = int((disp[row, col] - d_min) * subpixel)
                    itp_coeff[row, col] = cv[row, col, dsp]
                    if not np.isnan(cv[row, col, dsp]):
                        if (disp[row, col] != d_min) and (disp[row, col] != d_max):

                            sub_disp, sub_cost, valid = method(
                                [cv[row, col, dsp - 1], cv[row, col, dsp], cv[row, col, dsp + 1]],
                                disp[row, col],
                                measure)

                            disp[row, col] = sub_disp
                            itp_coeff[row, col] = sub_cost
                            mask[row, col] += valid
                        else:
                            # If Information: calculations stopped at the pixel step, sub-pixel interpolation did
                            # not succeed
                            mask[row, col] += cst.PANDORA_MSK_PIXEL_STOPPED_INTERPOLATION

        return itp_coeff, disp, mask

    @abstractmethod
    @njit(cache=True)
    def refinement_method(self, cost: np.ndarray, disp: float, measure: str) -> Tuple[float, float, int]:
        """
        Return the subpixel disparity and cost

        :param cost: cost of the values disp - 1, disp, disp + 1
        :type cost: 1D numpy array : [cost[disp -1], cost[disp], cost[disp + 1]]
        :param disp: the disparity
        :type disp: float
        :param measure: the type of measure used to create the cost volume
        :type measure: string = min | max
        :return: the refined disparity (disp + sub_disp), the refined cost and the state of the pixel( Information: \
        calculations stopped at the pixel step, sub-pixel interpolation did not succeed )
        :rtype: float, float, int
        """

    @staticmethod
    @njit(parallel=True)
    def loop_approximate_refinement(cv: np.ndarray, disp: np.ndarray, mask: np.ndarray, d_min: int, d_max: int,
                                    subpixel: int, measure: str, method: Callable[[np.ndarray, np.ndarray, np.ndarray,
                                                                                   np.ndarray, str], Tuple[
                                                                                      int, int, int]]) -> Tuple[
        np.ndarray, np.ndarray, np.ndarray]:
        """
         Apply for each pixels the refinement method on the right disparity map which was created with the approximate
          method : a diagonal search for the minimum on the left cost volume

        :param cv: the left cost volume
        :type cv: 3D numpy array (row, col, disp)
        :param disp: right disparity map
        :type disp: 2D numpy array (row, col)
        :param mask: right validity mask
        :type mask: 2D numpy array (row, col)
        :param d_min: minimal disparity
        :type d_min: int
        :param d_max: maximal disparity
        :type d_max: int
        :param subpixel: subpixel precision used to create the cost volume
        :type subpixel: int ( 1 | 2 | 4 )
        :param measure: the type of measure used to create the cost volume
        :type measure: string = min | max
        :param method: the refinement method
        :type method: function
        :return: the refine coefficient, the refine disparity map, and the validity mask
        :rtype: tuple(2D numpy array (row, col), 2D numpy array (row, col), 2D numpy array (row, col))
         """
        n_row, n_col, _ = cv.shape
        itp_coeff = np.zeros((n_row, n_col), dtype=np.float64)

        for row in prange(n_row):
            for col in prange(n_col):
                # No interpolation on invalid points
                if (mask[row, col] & cst.PANDORA_MSK_PIXEL_INVALID) != 0:
                    itp_coeff[row, col] = np.nan
                else:
                    # Conversion to numpy indexing
                    dsp = int((-disp[row, col] - d_min) * subpixel)
                    # Position of the best cost in the left cost volume is cv[r, diagonal, d]
                    diagonal = int(col + disp[row, col])
                    itp_coeff[row, col] = cv[row, diagonal, dsp]
                    if not np.isnan(cv[row, diagonal, dsp]):
                        if (disp[row, col] != -d_min) and (disp[row, col] != -d_max) and (diagonal != 0) and (
                                diagonal != (n_col - 1)):
                            # (1 * subpixel) because in fast mode, we can not have sub-pixel disparity for the right
                            # image.
                            # We therefore interpolate between pixel disparities
                            sub_disp, cost, valid = method([cv[row, diagonal - 1, dsp + (1 * subpixel)],
                                                            cv[row, diagonal, dsp],
                                                            cv[row, diagonal + 1, dsp - (1 * subpixel)]],
                                                           disp[row, col],
                                                           measure)

                            disp[row, col] = sub_disp
                            itp_coeff[row, col] = cost
                            mask[row, col] += valid
                        else:
                            # If Information: calculations stopped at the pixel step, sub-pixel interpolation did
                            # not succeed
                            mask[row, col] += cst.PANDORA_MSK_PIXEL_STOPPED_INTERPOLATION

        return itp_coeff, disp, mask



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions allowing to save the results and the configuration of Pandora pipeline.
"""

import errno
import json
import os
from typing import Dict, Tuple, List
import logging

import numpy as np
import rasterio.dtypes
import xarray as xr

from pandora.output_tree_design import get_out_dir, get_out_file_path
from pandora.img_tools import rasterio_open

def write_data_array(data_array: xr.DataArray, filename: str,
                     dtype: rasterio.dtypes = rasterio.dtypes.float32) -> None:
    """
    Write a xarray.DataArray in a tiff file

    :param data_array: data
    :type data_array: 2D xarray.DataArray (row, col) or 3D xarray.DataArray (row, col, indicator)
    :param filename:  output filename
    :type filename: string
    :param dtype: band types
    :type dtype: GDALDataType
    :return: None
    """
    if len(data_array.shape) == 2:
        row, col = data_array.shape
        with rasterio_open(filename, mode='w+', driver='GTiff', width=col, height=row, count=1,
                           dtype=dtype) as source_ds:
            source_ds.write(data_array.data, 1)

    else:
        row, col, depth = data_array.shape
        with rasterio_open(filename, mode='w+', driver='GTiff', width=col, height=row, count=depth,
                           dtype=dtype) as source_ds:
            for dsp in range(1, depth + 1):
                source_ds.write(data_array.data[:, :, dsp - 1], dsp)


def mkdir_p(path: str) -> None:
    """
    Create a directory without complaining if it already exists.
    :return: None
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # requires Python > 2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def save_results(left: xr.Dataset, right: xr.Dataset, output: str) -> None:
    """
    Save results in the output directory

    :param left: left dataset, which contains the variables :

        - disparity_map : the disparity map in the geometry of the left image 2D DataArray (row, col)
        - confidence_measure : the confidence measure in the geometry of the left image 3D DataArray \
        (row, col, indicator)
        - validity_mask : the validity mask in the geometry of the left image 2D DataArray (row, col)
    :type left: xr.Dataset
    :param right: right dataset. If there is no validation step, the right Dataset will be empty.If a validation step \
    is configured, the dataset will contain the variables :

        - disparity_map: the disparity map in the geometry of the right image 2D DataArray (row, col)
        - confidence_measure: the confidence in the geometry of the right image 3D DataArray (row, col, indicator)
        - validity_mask: the validity mask in the geometry of the left image 2D DataArray (row, col)
    :type right: xr.Dataset
    :param output: output directory
    :type output: string
    :return: None
    """
    # Create the output dir
    mkdir_p(output)

    # Save the left results
    write_data_array(left['disparity_map'], os.path.join(output, get_out_file_path('left_disparity.tif')))
    write_data_array(left['confidence_measure'], os.path.join(output, get_out_file_path('left_confidence_measure.tif')))
    write_data_array(left['validity_mask'], os.path.join(output, get_out_file_path('left_validity_mask.tif')),
                     dtype=rasterio.dtypes.uint16)

    # If a validation step is configured, save the right results
    if len(right.sizes) != 0:
        write_data_array(right['disparity_map'], os.path.join(output, get_out_file_path('right_disparity.tif')))
        write_data_array(right['confidence_measure'],
                         os.path.join(output, get_out_file_path('right_confidence_measure.tif')))
        write_data_array(right['validity_mask'], os.path.join(output, get_out_file_path('right_validity_mask.tif')),
                         dtype=rasterio.dtypes.uint16)


def sliding_window(base_array: np.array, shape: Tuple[int, int]) -> np.array:
    """
    Create a sliding window of using as_strided function : this function create a new a view (by manipulating
    data pointer) of the data array with a different shape. The new view pointing to the same memory block as
    data so it does not consume any additional memory.

    :param base_array: the 2D array through which slide the window
    :type base_array: np.array
    :param shape: shape of the sliding window
    :type shape: Tuple[int,int]

    :rtype: np.array
    """
    shp = (base_array.shape[0] - shape[0] + 1,) + (base_array.shape[1] - shape[1] + 1,) + shape
    strides = base_array.strides + base_array.strides
    return np.lib.stride_tricks.as_strided(base_array, shape=shp, strides=strides)


def save_config(output: str, user_cfg: Dict) -> None:
    """
    Save the user configuration in json file

    :param output: Path to output directory
    :type output: string
    :param user_cfg: user configuration
    :type user_cfg: dict
    :return: None
    """

    # Create the output dir
    mkdir_p(os.path.join(output, get_out_dir('config.json')))

    # Save user configuration in json file
    with open(os.path.join(output, get_out_file_path('config.json')), 'w') as file_:
        json.dump(user_cfg, file_, indent=2)


def is_method(string_method: str, methods: List[str]) -> bool:
    """
    Test if string_method is a method in methods

    :param string_method: String to test
    :type string_method: string
    :param methods: list of available methods
    :type methods: list of strings
    :returns: True if string_method a method and False otherwise
    :rtype: bool
    """

    if string_method in methods:
        return True

    logging.error('% is not in available methods : ', string_method + ', '.join(methods))
    return False



# pylint: disable=missing-module-docstring
#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora_pandora
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions allowing to check the configuration given to Pandora pipeline.
"""

import copy
import json
import logging
import sys
from collections.abc import Mapping
from typing import Dict, Union, List, Tuple

import numpy as np
from json_checker import Checker, Or, And

from pandora.state_machine import PandoraMachine
from pandora.img_tools import rasterio_open

from pandora import multiscale

def rasterio_can_open_mandatory(file_: str) -> bool:
    """
    Test if file can be open by rasterio

    :param file_: File to test
    :type file_: string
    :returns: True if rasterio can open file and False otherwise
    :rtype: bool
    """

    try:
        rasterio_open(file_)
        return True
    except Exception as exc:
        logging.warning('Impossible to read file %: %', file_, exc)
        return False


def rasterio_can_open(file_: str) -> bool:
    """
    Test if file can be open by rasterio

    :param file_: File to test
    :type file_: string
    :returns: True if rasterio can open file and False otherwise
    :rtype: bool
    """

    if file_ == 'none' or file_ is None:
        return True

    return rasterio_can_open_mandatory(file_)


def check_images(img_left: str, img_right: str, msk_left: str, msk_right: str) -> None:
    """
    Check the images

    :param img_left: path to the left image
    :type img_left: string
    :param img_right: path to the right image
    :type img_right: string
    :param msk_left: path to the mask of the left image
    :type msk_left: string
    :param msk_right: path to the mask of the right image
    :type msk_right: string
    :return: None
    """
    # verify that the images have 1 channel
    left_ = rasterio_open(img_left)
    if left_.count != 1:
        logging.error('The input images must be 1-channel grayscale images')
        sys.exit(1)
    right_ = rasterio_open(img_right)
    if right_.count != 1:
        logging.error('The input images must be 1-channel grayscale images')
        sys.exit(1)

    # verify that the images have the same size
    if (left_.width != right_.width) or \
            (left_.height != right_.height):
        logging.error('Images must have the same size')
        sys.exit(1)

    # verify that image and mask have the same size
    if msk_left is not None:
        msk_ = rasterio_open(msk_left)
        if (left_.width != msk_.width) or \
                (left_.height != msk_.height):
            logging.error('Image and masks must have the same size')
            sys.exit(1)

    # verify that image and mask have the same size
    if msk_right is not None:
        msk_ = rasterio_open(msk_right)
        # verify that the image and mask have the same size
        if (right_.width != msk_.width) or \
                (right_.height != msk_.height):
            logging.error('Image and masks must have the same size')
            sys.exit(1)


def check_disparities(disp_min: Union[int, str, None], disp_max: Union[int, str, None],
                      right_disp_min: Union[str, None], right_disp_max: Union[str, None],
                      img_left: str) -> None:
    """
    Check left and right disparities.

    :param disp_min: minimal disparity
    :type disp_min: int or str or None
    :param disp_max: maximal disparity
    :type disp_max: int or str or None
    :param right_disp_min: right minimal disparity
    :type right_disp_min: str or None
    :param right_disp_max: right maximal disparity
    :type right_disp_max: str or None
    :param img_left: path to the left image
    :type img_left: str
    :return: None
    """
    # --- Check left disparities
    # left disparity are integers
    if isinstance(disp_min, int) and isinstance(disp_max, int):
        if disp_max < disp_min:
            logging.error('Disp_max must be bigger than Disp_min')
            sys.exit(1)

    # left disparity are grids
    elif (isinstance(disp_min, str)) and (isinstance(disp_max, str)):
        # Load an image to compare the grid size
        img_left_ = rasterio_open(img_left)

        disp_min_ = rasterio_open(disp_min)
        dmin = disp_min_.read(1)
        disp_max_ = rasterio_open(disp_max)
        dmax = disp_max_.read(1)

        # check that disparity grids is a 1-channel grid
        if (disp_min_.count != 1) or (disp_max_.count != 1):
            logging.error('Disparity grids must be a 1-channel grid')
            sys.exit(1)

        # check that disp_min has the same size as the image
        if (disp_min_.width != img_left_.width) or (disp_min_.height != img_left_.height) \
                or (disp_max_.width != img_left_.width) or (disp_max_.height != img_left_.height):
            logging.error('Disparity grids and image must have the same size')
            sys.exit(1)

        if (dmax < dmin).any():
            logging.error('Disp_max must be bigger than Disp_min')
            sys.exit(1)

    # --- Check right disparities
    if (isinstance(right_disp_min, str)) and (isinstance(right_disp_max, str)):
        # Load an image to compare the grid size
        img_left_ = rasterio_open(img_left)

        disp_min_ = rasterio_open(disp_min)
        dmin = disp_min_.read(1)
        disp_max_ = rasterio_open(disp_max)
        dmax = disp_max_.read(1)

        # check that disparity grids is a 1-channel grid
        if (disp_min_.count != 1) or (disp_max_.count != 1):
            logging.error('Disparity grids must be a 1-channel grid')
            sys.exit(1)

        # check that disp_min has the same size as the image
        if (disp_min_.width != img_left_.width) or (disp_min_.height != img_left_.height) \
                or (disp_max_.width != img_left_.width) or (disp_max_.height != img_left_.height):
            logging.error('Disparity grids and image must have the same size')
            sys.exit(1)

        if (dmax < dmin).any():
            logging.error('Disp_max must be bigger than Disp_min')
            sys.exit(1)


def get_config_pipeline(user_cfg: Dict[str, dict]) -> Dict[str, dict]:
    """
    Get the pipeline configuration

    :param user_cfg: user configuration
    :type user_cfg: dict
    :return: cfg: partial configuration
    :rtype: cfg: dict
    """
    cfg = {}

    if 'pipeline' in user_cfg:
        cfg['pipeline'] = user_cfg['pipeline']

    return cfg


def get_config_input(user_cfg: Dict[str, dict]) -> Dict[str, dict]:
    """
    Get the input configuration

    :param user_cfg: user configuration
    :type user_cfg: dict
    :return cfg: partial configuration
    :rtype cfg: dict
    """

    cfg = {}

    if 'input' in user_cfg:
        cfg['input'] = user_cfg['input']

    return cfg


def check_pipeline_section(user_cfg: Dict[str, dict], pandora_machine: PandoraMachine) -> Dict[str, dict]:
    """
    Check if the pipeline is correct by
    - Checking the sequence of steps according to the machine transitions
    - Checking parameters, define in dictionary, of each Pandora step

    :param user_cfg: pipeline user configuration
    :type user_cfg: dict
    :param pandora_machine: instance of PandoraMachine
    :type pandora_machine: PandoraMachine object
    :return: cfg: pipeline configuration
    :rtype: cfg: dict
    """
    # Check if user configuration pipeline is compatible with transitions/states of pandora machine.
    cfg = update_conf(default_short_configuration_pipeline, user_cfg)
    pandora_machine.check_conf(cfg['pipeline'])

    cfg = update_conf(cfg, pandora_machine.pipeline_cfg)

    configuration_schema = {
        'pipeline': dict
    }

    checker = Checker(configuration_schema)
    checker.validate(cfg)

    return cfg


def check_input_section(user_cfg: Dict[str, dict]) -> Dict[str, dict]:
    """
    Complete and check if the dictionary is correct

    :param user_cfg: user configuration
    :type user_cfg: dict
    :return: cfg: global configuration
    :rtype: cfg: dict
    """
    # Add missing steps and inputs defaults values in user_cfg
    cfg = update_conf(default_short_configuration_input, user_cfg)

    # Disparity can be integer type, or string type (path to the disparity grid)
    # If the left disparity is string type, right disparity must be string type or none
    # if the left disparity is integer type, right disparity must be none
    if isinstance(cfg['input']['disp_min'], int):
        input_configuration_schema.update(input_configuration_schema_integer_disparity)
    else:
        if isinstance(cfg['input']['disp_min_right'], str):
            input_configuration_schema.update(input_configuration_schema_left_disparity_grids_right_grids)
        else:
            input_configuration_schema.update(input_configuration_schema_left_disparity_grids_right_none)

    # check schema
    configuration_schema = {
        'input': input_configuration_schema
    }

    checker = Checker(configuration_schema)
    checker.validate(cfg)

    # custom checking

    # check left and right disparities
    check_disparities(cfg['input']['disp_min'], cfg['input']['disp_max'], cfg['input']['disp_min_right'],
                      cfg['input']['disp_max_right'], cfg['input']['img_left'])

    check_images(cfg['input']['img_left'], cfg['input']['img_right'], cfg['input']['left_mask'],
                 cfg['input']['right_mask'])

    return cfg


def check_conf(user_cfg: Dict[str, dict], pandora_machine: PandoraMachine) -> dict:
    """
    Complete and check if the dictionary is correct

    :param user_cfg: user configuration
    :type user_cfg: dict
    :param pandora_machine: instance of PandoraMachine
    :type pandora_machine: PandoraMachine
    :return: cfg: global configuration
    :rtype: cfg: dict
    """

    # check input
    user_cfg_input = get_config_input(user_cfg)
    cfg_input = check_input_section(user_cfg_input)

    # check pipeline
    user_cfg_pipeline = get_config_pipeline(user_cfg)
    cfg_pipeline = check_pipeline_section(user_cfg_pipeline, pandora_machine)

    # If left disparities are grids of disparity and the right disparities are none, the cross-checking
    # method cannot be used

    if (isinstance(cfg_input['input']['disp_min'], str)) and (cfg_input['input']['disp_min_right'] is None) and \
            ('validation' in cfg_pipeline['pipeline']):
        logging.error('The cross-checking step cannot be processed if disp_min, disp_max are paths to the left '
                      'disparity grids and disp_right_min, disp_right_max are none.')
        sys.exit(1)

    if (isinstance(cfg_input['input']['disp_min'], str) or isinstance(cfg_input['input']['disp_min_right'], str) or \
        (isinstance(cfg_input['input']['disp_max'], str)) or isinstance(cfg_input['input']['disp_max_right'], str)) \
            and ('multiscale' in cfg_pipeline['pipeline']):
        logging.error('Multiscale processing does not accept input disparity grids.')
        sys.exit(1)

    # concatenate updated config
    cfg = concat_conf([cfg_input, cfg_pipeline])

    return cfg


def concat_conf(cfg_list: List[Dict[str, dict]]) -> Dict[str, dict]:
    """
    Concatenate dictionaries

    :param cfg_list: list of configurations
    :type cfg_list: List of dict
    :return: cfg: global configuration
    :rtype: cfg: dict
    """
    # concatenate updated config
    cfg = {}
    for conf in cfg_list:
        cfg.update(conf)

    return cfg


def read_multiscale_params(cfg: Dict[str, dict]) -> Tuple[int, int]:
    """
    Returns the multiscale parameters

    :param cfg: configuration
    :type cfg: dict
    :return:
        - num_scales: number of scales
        - scale_factor: factor by which each coarser layer is downsampled
    :rtype: tuple(int, int )
    """

    if 'multiscale' in cfg:
        # Multiscale processing in conf
        multiscale_ = multiscale.AbstractMultiscale(**cfg['multiscale'])

        num_scales = multiscale_.cfg['num_scales']
        scale_factor = multiscale_.cfg['scale_factor']
    else:
        # No multiscale selected
        num_scales = 1
        scale_factor = 1
    return num_scales, scale_factor

input_configuration_schema = {
    'img_left': And(str, rasterio_can_open_mandatory),
    'img_right': And(str, rasterio_can_open_mandatory),
    'nodata_left': Or(int, lambda input: np.isnan(input)),
    'nodata_right': Or(int, lambda input: np.isnan(input)),
    'left_mask': And(Or(str, lambda input: input is None), rasterio_can_open),
    'right_mask': And(Or(str, lambda input: input is None), rasterio_can_open),
    'left_classif': And(Or(str, lambda x: x is None), rasterio_can_open),
    'right_classif': And(Or(str, lambda x: x is None), rasterio_can_open),
    'left_segm': And(Or(str, lambda x: x is None), rasterio_can_open),
    'right_segm': And(Or(str, lambda x: x is None), rasterio_can_open)

}

# Input configuration when disparity is integer
input_configuration_schema_integer_disparity = {
    'disp_min': int,
    'disp_max': int,
    'disp_min_right': (lambda input: input is None),
    'disp_max_right': (lambda input: input is None)
}

# Input configuration when left disparity is a grid, and right not provided
input_configuration_schema_left_disparity_grids_right_none = {
    'disp_min': And(str, rasterio_can_open),
    'disp_max': And(str, rasterio_can_open),
    'disp_min_right': (lambda input: input is None),
    'disp_max_right': (lambda input: input is None)
}

# Input configuration when left and right disparity are grids
input_configuration_schema_left_disparity_grids_right_grids = {
    'disp_min': And(str, rasterio_can_open),
    'disp_max': And(str, rasterio_can_open),
    'disp_min_right': And(str, rasterio_can_open),
    'disp_max_right': And(str, rasterio_can_open)
}


default_short_configuration_input = {
    'input': {
        'nodata_left': -9999,
        'nodata_right': -9999,
        'left_mask': None,
        'right_mask': None,
        'left_classif': None,
        'right_classif': None,
        'left_segm': None,
        'right_segm': None,
        'disp_min_right': None,
        'disp_max_right': None
    }
}

default_short_configuration_pipeline = {
    'pipeline':
        {
            'right_disp_map': {
                'method': 'none'
            }
        }
}

default_short_configuration = concat_conf([default_short_configuration_input,
                                           default_short_configuration_pipeline])

def read_config_file(config_file: str) -> Dict[str, dict]:
    """
    Read a json configuration file

    :param config_file: path to a json file containing the algorithm parameters
    :type config_file: string
    :return user_cfg: configuration dictionary
    :rtype: dict
    """
    with open(config_file, 'r') as file_:
        user_cfg = json.load(file_)
    return user_cfg


def update_conf(def_cfg: Dict[str, dict], user_cfg: Dict[str, dict]) -> Dict[str, dict]:
    """
    Update the default configuration with the user configuration,

    :param def_cfg: default configuration
    :type def_cfg: dict
    :param user_cfg: user configuration
    :type user_cfg: dict
    :return: the user and default configuration
    :rtype: dict
    """
    config = copy.deepcopy(def_cfg)
    for key, value in user_cfg.items():
        if isinstance(value, Mapping):
            config[key] = update_conf(config.get(key, {}), value)
        else:
            if value == 'NaN':
                value = np.nan
            config[key] = value
    return config



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to SAD and SSD methods used in the cost volume measure step.
"""

from typing import Dict, Union, Tuple

import numpy as np
import xarray as xr
from json_checker import Checker, And

from pandora.img_tools import shift_right_img
from pandora import common
from pandora.matching_cost import matching_cost


@matching_cost.AbstractMatchingCost.register_subclass('sad', 'ssd')
class SadSsd(matching_cost.AbstractMatchingCost):
    """
    SadSsd class allows to compute the cost volume
    """
    # Default configuration, do not change these values
    _WINDOW_SIZE = 5
    _SUBPIX = 1

    def __init__(self, **cfg: Union[str, int]) -> None:
        """
        :param cfg: optional configuration,  {'matching_cost_method': value, 'window_size': value, 'subpix': value}
        :type cfg: dict
        :return: None
        """
        self.cfg = self.check_conf(**cfg)
        self._method = str(self.cfg['matching_cost_method'])
        self._window_size = self.cfg['window_size']
        self._subpix = self.cfg['subpix']
        self._pixel_wise_methods = {'sad': self.ad_cost, 'ssd': self.sd_cost}

    def check_conf(self, **cfg: Union[str, int]) -> Dict[str, Union[str, int]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: matching cost configuration
        :type cfg: dict
        :return cfg: matching cost configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the conf
        if 'window_size' not in cfg:
            cfg['window_size'] = self._WINDOW_SIZE
        if 'subpix' not in cfg:
            cfg['subpix'] = self._SUBPIX

        schema = {
            'matching_cost_method': And(str, lambda input: common.is_method(input, ['ssd', 'sad'])),
            'window_size': And(int, lambda input: input > 0 and (input % 2) != 0),
            'subpix': And(int, lambda input: input in (1, 2, 4))
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the matching cost method
        :return: None
        """
        print(str(self._method) + ' similarity measure')

    def compute_cost_volume(self, img_left: xr.Dataset, img_right: xr.Dataset, disp_min: int,
                            disp_max: int) -> xr.Dataset:
        """
        Computes the cost volume for a pair of images

        :param img_left: left Dataset image
        :type img_left:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :param img_right: right Dataset image
        :type img_right:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :param disp_min: minimum disparity
        :type disp_min: int
        :param disp_max: maximum disparity
        :type disp_max: int
        :return: the cost volume dataset
        :rtype:
            xarray.Dataset, with the data variables:
                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        """
        # Contains the shifted right images
        img_right_shift = shift_right_img(img_right, self._subpix)

        # Computes the maximal cost of the cost volume
        min_left = np.amin(img_left['im'].data)
        max_left = np.amax(img_left['im'].data)

        min_right = np.amin(img_right['im'].data)
        max_right = np.amax(img_right['im'].data)
        cmax = None

        if self._method == 'sad':
            # Maximal cost of the cost volume with sad measure
            cmax = int(max(abs(max_left - min_right), abs(max_right - min_left)) * (self._window_size ** 2))
        if self._method == 'ssd':
            # Maximal cost of the cost volume with ssd measure
            cmax = int(max(abs(max_left - min_right) ** 2, abs(max_right - min_left) ** 2) * (self._window_size ** 2))
        offset_row_col = int((self._window_size - 1) / 2)
        metadata = {'measure': self._method, 'subpixel': self._subpix,
                    'offset_row_col': offset_row_col, 'window_size': self._window_size,
                    'type_measure': 'min', 'cmax': cmax}

        # Disparity range # pylint: disable=undefined-variable
        if self._subpix == 1:
            disparity_range = np.arange(disp_min, disp_max + 1)
        else:
            disparity_range = np.arange(disp_min, disp_max, step=1 / float(self._subpix))
            disparity_range = np.append(disparity_range, [disp_max])

        # Allocate the numpy cost volume cv = (disp, col, row), for efficient memory management
        # If offset , over allocate the cost volume by adding 2 * offset on row and col dimension
        # The following computation will reduce the dimension during the pixel wise aggregation so the final
        # cv dimension will be equal to the correct one.
        if offset_row_col != 0:
            cv_enlarge = np.zeros((len(disparity_range), img_left['im'].shape[1] + 2 * offset_row_col,
                                   img_left['im'].shape[0] + 2 * offset_row_col), dtype=np.float32)
            cv_enlarge += np.nan
            cv = cv_enlarge[:, offset_row_col:-offset_row_col, offset_row_col: -offset_row_col]
        else:
            cv = np.zeros((len(disparity_range), img_left['im'].shape[1], img_left['im'].shape[0] ), dtype=np.float32)
            cv += np.nan

        # Giving the 2 images, the matching cost will be calculated as :
        #                 1, 1, 1                2, 5, 6
        #                 2, 1, 4                1, 1, 3
        #                 1, 3, 1                6, 1, 2
        #
        # for disp = 0, the costvolume will be
        #                        (1 - 2) (1 - 5) (1 - 6)
        #                        (2 - 1) (1 - 1) (4 - 3)
        #                        (1 - 6) (3 - 1) (1 - 2)
        # , we make the difference between the two full images
        #
        # for disp = -1, compute the difference between
        #                    1, 1                   2, 5
        #                    1, 4                   1, 1
        #                    3, 1                   6, 1
        # , the costvolume will be
        #                        nan  (1-2) (1-5)
        #                        nan  (1-1) (4-1)
        #                        nan  (3-6) (1-1)
        # , nan correspond to the first column of the left image that cannot be calculated,
        #  because there is no corresponding column for the disparity -1.
        #
        # In the following loop, the tuples point_p,
        # point_q describe to which part of the image the difference will be applied,
        # for disp = 0, we compute the difference on the whole images so point_p=(:,0:3) et point_q=(:,0:3),
        # for disp = -1, we use only a part of the images so point_p=(:,1:3) et point_q=(:,0:2)

        # Computes the matching cost
        # In the loop, cv is of shape (disp, col, row) and images / masks of shape (row, col)
        # np.swapaxes allow to interchange row and col in images and masks
        for disp in disparity_range:
            i_right = int((disp % 1) * self._subpix)
            point_p, point_q = self.point_interval(img_left, img_right_shift[i_right], disp)
            dsp = int((disp - disp_min) * self._subpix)

            cv[dsp, point_p[0]:point_p[1], :] = \
                np.swapaxes(
                    self._pixel_wise_methods[self._method](point_p, point_q, img_left, img_right_shift[i_right]), 0, 1)

        # Pixel wise aggregation modifies border values so it is important to reconvert to nan values
        if offset_row_col != 0:
            cv = self.pixel_wise_aggregation(cv_enlarge.data)
            cv = np.swapaxes(cv, 0, 2)
            cv[:offset_row_col, :, :] = np.nan
            cv[-offset_row_col:, :, ] = np.nan
            cv[:, :offset_row_col, :] = np.nan
            cv[:, -offset_row_col:,:] = np.nan
        else:
            cv = self.pixel_wise_aggregation(cv.data)
            cv = np.swapaxes(cv, 0, 2)

        # Create the xarray.DataSet that will contain the cv of dimensions (row, col, disp)
        cv = self.allocate_costvolume(img_left, self._subpix, disp_min, disp_max, self._window_size, metadata,
                                      cv)

        return cv

    @staticmethod
    def ad_cost(point_p: Tuple[int, int], point_q: Tuple[int, int], img_left: xr.Dataset,
                img_right: xr.Dataset) -> np.ndarray:
        """
        Computes the absolute difference

        :param point_p: Point interval, in the left image, over which the squared difference will be applied
        :type point_p: tuple
        :param point_q: Point interval, in the right image, over which the squared difference will be applied
        :type point_q: tuple
        :param img_left: left Dataset image
        :type img_left:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :param img_right: right Dataset image
        :type img_right:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :return: the absolute difference pixel-wise between elements in the interval
        :rtype: numpy array
        """
        return abs(img_left['im'].data[:, point_p[0]:point_p[1]] - img_right['im'].data[:, point_q[0]:point_q[1]])

    @staticmethod
    def sd_cost(point_p: Tuple, point_q: Tuple, img_left: xr.Dataset, img_right: xr.Dataset) -> np.ndarray:
        """
        Computes the square difference

        :param point_p: Point interval, in the left image, over which the squared difference will be applied
        :type point_p: tuple
        :param point_q: Point interval, in the right image, over which the squared difference will be applied
        :type point_q: tuple
        :param img_left: left Dataset image
        :type img_left:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :param img_right: right Dataset image
        :type img_right:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :return: the squared difference pixel-wise between elements in the interval
        :rtype: numpy array
        """
        return (img_left['im'].data[:, point_p[0]:point_p[1]] - img_right['im'].data[:, point_q[0]:point_q[1]]) ** 2

    def pixel_wise_aggregation(self, cost_volume: np.ndarray) -> np.ndarray:
        """
        Summing pixel wise matching cost over square windows

         :param cost_volume: the cost volume
         :type cost_volume: numpy array 3D (disp, col, row)
         :return: the cost volume aggregated
         :rtype: numpy array 3D ( disp, col, row)
        """
        nb_disp, nx_, ny_ = cost_volume.shape

        # Create a sliding window of using as_strided function : this function create a new a view (by manipulating
        # data pointer) of the cost_volume array with a different shape. The new view pointing to the same memory block
        # as cost_volume so it does not consume any additional memory.
        str_disp, str_col, str_row = cost_volume.strides

        shape_windows = (
            self._window_size, self._window_size, nb_disp, nx_ - (self._window_size - 1), ny_ - (self._window_size - 1))
        strides_windows = (str_row, str_col, str_disp, str_col, str_row)
        aggregation_window = np.lib.stride_tricks.as_strided(cost_volume, shape_windows, strides_windows,
                                                             writeable=False)
        cost_volume = np.sum(aggregation_window, (0, 1))
        return cost_volume



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the median filter used to filter the disparity map.
"""
import warnings
from typing import Dict, Union, cast

import numpy as np
import xarray as xr
from json_checker import Checker, And

import pandora.constants as cst
from . import filter # pylint: disable= redefined-builtin
from ..common import sliding_window


@filter.AbstractFilter.register_subclass('median')
class MedianFilter(filter.AbstractFilter):
    """
    MedianFilter class allows to perform the filtering step
    """
    # Default configuration, do not change this value
    _FILTER_SIZE = 3

    def __init__(self, **cfg: Union[str, int]):
        """
        :param cfg: optional configuration, {'filter_size': value}
        :type cfg: dictionary
        """
        self.cfg = self.check_conf(**cfg)
        self._filter_size = cast(int, self.cfg['filter_size'])

    def check_conf(self, **cfg: Union[str, int]) -> Dict[str, Union[str, int]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: filter configuration
        :type cfg: dict
        :return cfg: filter configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'filter_size' not in cfg:
            cfg['filter_size'] = self._FILTER_SIZE

        schema = {
            'filter_method': And(str, lambda input: 'median'),
            'filter_size': And(int, lambda input: input >= 1 and input % 2 != 0)
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self):
        """
        Describes the filtering method
        """
        print('Median filter description')

    def filter_disparity(self, disp: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None,
                         cv: xr.Dataset = None) -> None:
        """
        Apply a median filter on valid pixels.
        Invalid pixels are not filtered. If a valid pixel contains an invalid pixel in its filter, the invalid pixel is
        ignored for the calculation of the median.

        :param disp: the disparity map dataset with the variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
                - validity_mask 2D xarray.DataArray (row, col)
        :type disp: xarray.Dataset
        :param img_left: left Dataset image
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image
        :type img_right: xarray.Dataset
        :param cv: cost volume dataset
        :type cv: xarray.Dataset
        :return: None
        """
        # Invalid pixels are nan
        masked_data = disp['disparity_map'].copy(deep=True).data
        masked_data[np.where((disp['validity_mask'].data & cst.PANDORA_MSK_PIXEL_INVALID) != 0)] = np.nan

        valid = np.isfinite(masked_data)
        disp_median = self.median_filter(masked_data)

        disp['disparity_map'].data[valid] = disp_median[valid]
        disp.attrs['filter'] = 'median'
        del disp_median, masked_data,

    def median_filter(self, data) -> np.ndarray:
        """
        Apply median filter on valid pixels (pixels that are not nan).
        Invalid pixels are not filtered. If a valid pixel contains an invalid pixel in its filter, the invalid pixel is
        ignored for the calculation of the median.

        :param data: input data to be filtered
        :type data: 2D np.array (row, col)
        :return: The filtered array
        :rtype: 2D np.array(row, col)
        """
        data_median = np.copy(data)
        invalid = np.isnan(data_median)
        ny_, nx_ = data.shape

        aggregation_window = sliding_window(data, (self._filter_size, self._filter_size))

        radius = int(self._filter_size / 2)

        # To reduce memory, the data array is split (along the row axis) into multiple sub-arrays with a step of 100
        chunk_size = 100
        disp_chunked_y = np.array_split(aggregation_window, np.arange(chunk_size, ny_, chunk_size), axis=0)
        y_begin = radius

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
            # numpy.nanmedian : Compute the median along the specified axis, while ignoring NaNs (i.e if valid pixel
            # contains an invalid pixel in its filter, the invalid pixel is ignored because invalid pixels are nan )
            for col, disp_y in enumerate(disp_chunked_y): # pylint: disable= unused-variable
                # To reduce memory, the data array is split (along the col axis) into multiple sub-arrays,
                # with a step of 100
                disp_chunked_x = np.array_split(disp_y, np.arange(chunk_size, nx_, chunk_size), axis=1)
                x_begin = radius

                for rox, disp_x in enumerate(disp_chunked_x): # pylint: disable= unused-variable
                    y_end = y_begin + disp_y.shape[0]
                    x_end = x_begin + disp_x.shape[1]
                    data_median[y_begin:y_end, x_begin:x_end] = np.nanmedian(disp_x, axis=(2, 3))
                    x_begin += disp_x.shape[1]

                y_begin += disp_y.shape[0]

        del aggregation_window

        data_median[invalid] = np.nan
        return data_median



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the output tree structure.
"""

import os

OTD = {
    'left_disparity.tif': '.',
    'right_disparity.tif': '.',
    'left_confidence_measure.tif': '.',
    'right_confidence_measure.tif': '.',
    'left_validity_mask.tif': '.',
    'right_validity_mask.tif': '.',

    # Configuration
    'config.json': './cfg',
    'command_line.txt': './cfg'
}


def get_out_dir(key: str) -> str:
    """
    Return the output directory
    :param key: output product
    """
    return OTD[key]


def get_out_file_path(key: str) -> str:
    """
    Return the output file path
    :param key: output product
    """
    return os.path.join(get_out_dir(key), key)



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains classes and functions associated to the interpolation of the disparity map for the validation step.
"""

import logging
import math
from abc import ABCMeta, abstractmethod
from typing import Tuple

import numpy as np
import xarray as xr
from numba import njit

import pandora.constants as cst


class AbstractInterpolation():
    """
    Abstract Interpolation class
    """
    __metaclass__ = ABCMeta

    interpolation_methods_avail = {}

    def __new__(cls, **cfg: dict) -> object:
        """
        Return the plugin associated with the interpolated_disparity given in the configuration

        :param cfg: configuration {'interpolated_disparity': value}
        :type cfg: dictionary
        """
        if cls is AbstractInterpolation:
            if isinstance(cfg['interpolated_disparity'], str):
                try:
                    return super(AbstractInterpolation, cls).__new__(
                        cls.interpolation_methods_avail[cfg['interpolated_disparity']])
                except KeyError:
                    logging.error('No interpolation method named % supported', cfg['interpolated_disparity'])
                    raise KeyError
            else:
                if isinstance(cfg['interpolated_disparity'], unicode): # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractInterpolation, cls).__new__(
                            cls.interpolation_methods_avail[cfg['interpolated_disparity'].encode('utf-8')])
                    except KeyError:
                        logging.error('No interpolation method named % supported', cfg['interpolated_disparity'])
                        raise KeyError
        else:
            return super(AbstractInterpolation, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.interpolation_methods_avail[short_name] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self) -> None:
        """
        Describes the disparity interpolation method for the validation step
        :return: None
        """
        print('Disparity interpolation method description for the validation step')

    @abstractmethod
    def interpolated_disparity(self, left: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None,
                               cv: xr.Dataset = None) -> None:
        """
        Interpolation of the left disparity map to resolve occlusion and mismatch conflicts.

        :param left: left Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type left: xarray.Dataset
        :param img_left: left Datset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost_volume Dataset with the variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: None
        """


@AbstractInterpolation.register_subclass('mc-cnn')
class McCnnInterpolation(AbstractInterpolation):
    """
        McCnnInterpolation class allows to perform the interpolation of the disparity map
        """

    def __init__(self, **cfg: dict) -> None:
        """
        :param cfg: optional configuration, {}
        :type cfg: dictionary
        :return: None
        """
        self.check_config(**cfg)

    def check_config(self, **cfg: dict) -> None:
        """
        Check and update the configuration

        :param cfg: optional configuration, {}
        :type cfg: dictionary
        :return: None
        """
        # No optional configuration

    def desc(self) -> None:
        """
        Describes the disparity interpolation method
        :return: None
        """
        print('MC-CNN interpolation method')

    def interpolated_disparity(self, left: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None,
                               cv: xr.Dataset = None) -> None:
        """
        Interpolation of the left disparity map to resolve occlusion and mismatch conflicts.

        :param left: left Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type left: xarray.Dataset
        :param img_left: left Datset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost_volume Dataset with the variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: None
        """

        left['disparity_map'].data, left['validity_mask'].data = \
            self.interpolate_occlusion_mc_cnn(left['disparity_map'].data, left['validity_mask'].data)
        left['disparity_map'].data, left['validity_mask'].data = \
            self.interpolate_mismatch_mc_cnn(left['disparity_map'].data, left['validity_mask'].data)

        left.attrs['interpolated_disparity'] = 'mc-cnn'

    @staticmethod
    @njit()
    def interpolate_occlusion_mc_cnn(disp: np.ndarray, valid: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Interpolation of the left disparity map to resolve occlusion conflicts.
        Interpolate occlusion by moving left until
        we find a position labeled correct.

        Žbontar, J., & LeCun, Y. (2016). Stereo matching by training a convolutional neural network to compare image
        patches. The journal of machine learning research, 17(1), 2287-2318.

        :param disp: disparity map
        :type disp: 2D np.array (row, col)
        :param valid: validity mask
        :type valid: 2D np.array (row, col)
        :return: the interpolate left disparity map, with the validity mask update :

            - If out & MSK_PIXEL_FILLED_OCCLUSION != 0 : Invalid pixel : filled occlusion
        :rtype: tuple(2D np.array (row, col), 2D np.array (row, col))
        """
        # Output disparity map and validity mask
        out_disp = np.copy(disp)
        out_val = np.copy(valid)

        ncol, nrow = disp.shape
        for col in range(ncol):
            for row in range(nrow):
                # Occlusion
                if (valid[col, row] & cst.PANDORA_MSK_PIXEL_OCCLUSION) != 0:
                    # interpolate occlusion by moving left until we find a position labeled correct

                    #  valid pixels mask
                    msk = (valid[col, 0:row + 1] & cst.PANDORA_MSK_PIXEL_INVALID) == 0
                    # Find the first valid pixel
                    msk = msk[::-1]
                    arg_valid = np.argmax(msk)

                    # If occlusions are still present :  interpolate occlusion by moving right until we find a position
                    # labeled correct
                    if arg_valid == 0:
                        # valid pixels mask
                        msk = (valid[col, row:] & cst.PANDORA_MSK_PIXEL_INVALID) == 0
                        # Find the first valid pixel
                        arg_valid = np.argmax(msk)

                        # Update the validity mask Information : filled occlusion
                        out_val[col, row] -= cst.PANDORA_MSK_PIXEL_OCCLUSION * msk[arg_valid]
                        out_val[col, row] += cst.PANDORA_MSK_PIXEL_FILLED_OCCLUSION * msk[arg_valid]
                        out_disp[col, row] = disp[col, row + arg_valid]
                    else:
                        # Update the validity mask : Information : filled occlusion
                        out_val[col, row] -= cst.PANDORA_MSK_PIXEL_OCCLUSION * msk[arg_valid]
                        out_val[col, row] += cst.PANDORA_MSK_PIXEL_FILLED_OCCLUSION * msk[arg_valid]
                        out_disp[col, row] = disp[col, row - arg_valid]

        return out_disp, out_val

    @staticmethod
    @njit()
    def interpolate_mismatch_mc_cnn(disp: np.ndarray, valid: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Interpolation of the left disparity map to resolve mismatch conflicts.
        Interpolate mismatch by finding the nearest
        correct pixels in 16 different directions and use the median of their disparities.

        Žbontar, J., & LeCun, Y. (2016). Stereo matching by training a convolutional neural network to compare image
        patches. The journal of machine learning research, 17(1), 2287-2318.

        :param disp: disparity map
        :type disp: 2D np.array (row, col)
        :param valid: validity mask
        :type valid: 2D np.array (row, col)
        :return: the interpolate left disparity map, with the validity mask update :

            - If out & MSK_PIXEL_FILLED_MISMATCH != 0 : Invalid pixel : filled mismatch
        :rtype: tuple(2D np.array (row, col), 2D np.array (row, col))
        """
        # Output disparity map and validity mask
        out_disp = np.copy(disp)
        out_val = np.copy(valid)

        ncol, nrow = disp.shape

        # 16 directions : [row, col]
        dirs = np.array([[0., 1.], [-0.5, 1.], [-1., 1.], [-1., 0.5], [-1., 0.], [-1., -0.5], [-1., -1.], [-0.5, -1.],
                         [0., -1.], [0.5, -1.], [1., -1.], [1., -0.5], [1., 0.], [1., 0.5], [1., 1.], [0.5, 1.]])

        # Maximum path length
        max_path_length = max(nrow, ncol)

        for col in range(ncol):
            for row in range(nrow):
                # Mismatch
                if (valid[col, row] & cst.PANDORA_MSK_PIXEL_MISMATCH) != 0:
                    interp_mismatched = np.zeros(16, dtype=np.float32)
                    # For each directions
                    for direction in range(16):
                        # Find the first valid pixel in the current path
                        for i in range(1, max_path_length):
                            tmp_row = row + int(dirs[direction][0] * i)
                            tmp_col = col + int(dirs[direction][1] * i)
                            tmp_row = math.floor(tmp_row)
                            tmp_col = math.floor(tmp_col)

                            # Edge of the image reached: there is no valid pixel in the current path
                            if (tmp_col < 0) | (tmp_col >= ncol) | (tmp_row < 0) | (tmp_row >= nrow):
                                interp_mismatched[direction] = np.nan
                                break

                            # First valid pixel
                            if (valid[tmp_col, tmp_row] & cst.PANDORA_MSK_PIXEL_INVALID) == 0:
                                interp_mismatched[direction] = disp[tmp_col, tmp_row]
                                break

                    # Median of the 16 pixels
                    out_disp[col, row] = np.nanmedian(interp_mismatched)
                    # Update the validity mask : Information : filled mismatch
                    out_val[col, row] -= cst.PANDORA_MSK_PIXEL_MISMATCH
                    out_val[col, row] += cst.PANDORA_MSK_PIXEL_FILLED_MISMATCH

        return out_disp, out_val


@AbstractInterpolation.register_subclass('sgm')
class SgmInterpolation(AbstractInterpolation):
    """
        SgmInterpolation class allows to perform the interpolation of the disparity map
        """

    def __init__(self, **cfg: dict) -> None:
        """
        :param cfg: optional configuration, {}
        :type cfg: dictionary
        :return: None
        """
        self.check_config(**cfg)

    def check_config(self, **cfg: dict) -> None:
        """
        Check and update the configuration

        :param cfg: optional configuration, {}
        :type cfg: dictionary
        :return: None
        """
        # No optional configuration

    def desc(self) -> None:
        """
        Describes the disparity interpolation method
        :return: None
        """
        print('SGM interpolation method')

    def interpolated_disparity(self, left: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None,
                               cv: xr.Dataset = None) -> None:
        """
        Interpolation of the left disparity map to resolve occlusion and mismatch conflicts.

        :param left: left Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type left: xarray.Dataset
        :param img_left: left Datset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost_volume Dataset with the variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: None
        """

        left['disparity_map'].data, left['validity_mask'].data = self.interpolate_mismatch_sgm(
            left['disparity_map'].data,
            left['validity_mask'].data)
        left['disparity_map'].data, left['validity_mask'].data = self.interpolate_occlusion_sgm(
            left['disparity_map'].data,
            left['validity_mask'].data)
        left.attrs['interpolated_disparity'] = 'sgm'

    def interpolate_occlusion_sgm(self, disp: np.ndarray, valid: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Interpolation of the left disparity map to resolve occlusion conflicts.
        Interpolate occlusion by moving by selecting
        the right lowest value along paths from 8 directions.

        HIRSCHMULLER, Heiko. Stereo processing by semiglobal matching and mutual information.
        IEEE Transactions on pattern analysis and machine intelligence, 2007, vol. 30, no 2, p. 328-341.

        :param disp: disparity map
        :type disp: 2D np.array (row, col)
        :param valid: validity mask
        :type valid: 2D np.array (row, col)
        :return: the interpolate left disparity map, with the validity mask update :

            - If out & MSK_PIXEL_FILLED_OCCLUSION != 0 : Invalid pixel : filled occlusion
        :rtype: : tuple(2D np.array (row, col), 2D np.array (row, col))
        """
        # Output disparity map and validity mask
        out_disp = np.copy(disp)
        out_val = np.copy(valid)

        # 8 directions : [row, col]
        dirs = np.array([[0, 1], [-1, 1], [-1, 0], [-1, -1], [0, -1], [1, -1], [1, 0], [1, 1]])

        ncol, nrow = disp.shape
        for col in range(ncol):
            for row in range(nrow):
                # Occlusion
                if (valid[col, row] & cst.PANDORA_MSK_PIXEL_OCCLUSION) != 0:
                    valid_neighbors = self.find_valid_neighbors(dirs, disp, valid, row, col)

                    # Returns the indices that would sort the absolute array
                    # The absolute value is used to search for the right value closest to 0
                    valid_neighbors_argsort = np.argsort(np.abs(valid_neighbors))

                    # right lowest value
                    out_disp[col, row] = valid_neighbors[valid_neighbors_argsort[1]]

                    # Update the validity mask : Information : filled occlusion
                    out_val[col, row] -= cst.PANDORA_MSK_PIXEL_OCCLUSION
                    out_val[col, row] += cst.PANDORA_MSK_PIXEL_FILLED_OCCLUSION
        return out_disp, out_val

    def interpolate_mismatch_sgm(self, disp: np.ndarray, valid: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Interpolation of the left disparity map to resolve mismatch conflicts. Interpolate mismatch by finding the
        nearest correct pixels in 8 different directions and use the median of their disparities.
        Mismatched pixel areas that are direct neighbors of occluded pixels are treated as occlusions.

        HIRSCHMULLER, Heiko. Stereo processing by semiglobal matching and mutual information.
        IEEE Transactions on pattern analysis and machine intelligence, 2007, vol. 30, no 2, p. 328-341.

        :param disp: disparity map
        :type disp: 2D np.array (row, col)
        :param valid: validity mask
        :type valid: 2D np.array (row, col)
        :return: the interpolate left disparity map, with the validity mask update :

            - If out & MSK_PIXEL_FILLED_MISMATCH != 0 : Invalid pixel : filled mismatch
        :rtype: tuple(2D np.array (row, col), 2D np.array (row, col))
        """
        # Output disparity map and validity mask
        out_disp = np.copy(disp)
        out_val = np.copy(valid)

        # 8 directions : [row, col]
        dirs = np.array([[0, 1], [-1, 1], [-1, 0], [-1, -1], [0, -1], [1, -1], [1, 0], [1, 1]])

        ncol, nrow = disp.shape
        for col in range(ncol):
            for row in range(nrow):
                # Mismatched
                if valid[col, row] & cst.PANDORA_MSK_PIXEL_MISMATCH != 0:

                    # Mismatched pixel areas that are direct neighbors of occluded pixels are treated as occlusions
                    if np.sum(valid[max(0, col - 1):min(ncol - 1, col + 1) + 1,
                              max(0, row - 1):min(nrow - 1, row + 1) + 1] &
                              cst.PANDORA_MSK_PIXEL_OCCLUSION) != 0:
                        out_val[col, row] -= cst.PANDORA_MSK_PIXEL_MISMATCH
                        out_val[col, row] += cst.PANDORA_MSK_PIXEL_OCCLUSION

                    else:
                        valid_neighbors = self.find_valid_neighbors(dirs, disp, valid, row, col)

                        # Median of the 8 pixels
                        out_disp[col, row] = np.nanmedian(valid_neighbors)
                        # Update the validity mask : Information : filled mismatch
                        out_val[col, row] -= cst.PANDORA_MSK_PIXEL_MISMATCH
                        out_val[col, row] += cst.PANDORA_MSK_PIXEL_FILLED_MISMATCH

        return out_disp, out_val

    @staticmethod
    @njit()
    def find_valid_neighbors(dirs: np.ndarray, disp: np.ndarray, valid: np.ndarray,
                             row: int, col: int):
        """
        :param dirs: directions
        :type dirs: 2D np.array (row, col)
        :param disp: disparity map
        :type disp: 2D np.array (row, col)
        :param valid: validity mask
        :type valid: 2D np.array (row, col)
        :param row: row current value
        :type row: int
        :param col: col current value
        :type col: int
        :return: valid neighbors
        :rtype: 2D np.array
        """
        ncol, nrow = disp.shape
        # Maximum path length
        max_path_length = max(nrow, ncol)
        # For each directions
        valid_neighbors = np.zeros(8, dtype=np.float32)
        for direction in range(8):
            # Find the first valid pixel in the current path
            tmp_row = row
            tmp_col = col
            for i in range(max_path_length):  # pylint: disable= unused-variable
                tmp_row += dirs[direction][0]
                tmp_col += dirs[direction][1]

                # Edge of the image reached: there is no valid pixel in the current path
                if (tmp_col < 0) | (tmp_col >= ncol) | (tmp_row < 0) | (tmp_row >= nrow):
                    valid_neighbors[direction] = np.nan
                    break
                # First valid pixel
                if (valid[tmp_col, tmp_row] & cst.PANDORA_MSK_PIXEL_INVALID) == 0:
                    valid_neighbors[direction] = disp[tmp_col, tmp_row]
                    break
        return valid_neighbors



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora_pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the multi-scale pyramid method.
"""

import warnings
from typing import Dict, Union, Tuple

import numpy as np
import xarray as xr
from json_checker import Checker, And
from scipy.ndimage.interpolation import zoom

from . import multiscale
from ..common import sliding_window


@multiscale.AbstractMultiscale.register_subclass('fixed_zoom_pyramid')
class FixedZoomPyramid(multiscale.AbstractMultiscale):
    """
    FixedZoomPyramid class, allows to perform the multiscale processing
    """
    _PYRAMID_NUM_SCALES = 2
    _PYRAMID_SCALE_FACTOR = 2
    _PYRAMID_MARGE = 1

    def __init__(self, **cfg: dict):
        """
        :param cfg: optional configuration, {  "num_scales": int, "scale_factor": int, "marge" int}
        :type cfg: dict
        """
        self.cfg = self.check_conf(**cfg)
        self._num_scales = self.cfg['num_scales']
        self._scale_factor = self.cfg['scale_factor']
        self._marge = self.cfg['marge']

    def check_conf(self, **cfg: Union[str, float, int]) -> Dict[str, Union[str, float, int]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: aggregation configuration
        :type cfg: dict
        :return cfg: aggregation configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'num_scales' not in cfg:
            cfg['num_scales'] = self._PYRAMID_NUM_SCALES
        if 'scale_factor' not in cfg:
            cfg['scale_factor'] = self._PYRAMID_SCALE_FACTOR
        if 'marge' not in cfg:
            cfg['marge'] = self._PYRAMID_MARGE

        schema = {
            'multiscale_method': And(str, lambda x: 'fixed_zoom_pyramid'),
            'num_scales': And(int, lambda x: x > 1),
            'scale_factor': And(int, lambda x: x > 1),
            'marge': And(int, lambda x: x >= 0),
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self):
        """
        Describes the aggregation method
        """
        print('FixedZoomPyramid method')

    def disparity_range(self, disp: xr.Dataset, disp_min: int, disp_max: int) -> \
            Tuple[np.array, np.array]:
        """
        Disparity range computation by seeking the max and min values in the window.
        Invalid disparities are given the full disparity range

        :param disp: the disparity dataset
        :type disp: xarray.Dataset with the data variables :
                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :param disp_min: absolute min disparity
        :type disp_min: int
        :param disp_max: absolute max disparity
        :type disp_max: int
        :return: Two np.darray :
                - disp_min_range : minimum disparity value for all pixels.
                - disp_max_range : maximum disparity value for all pixels.

        :rtype: tuple (np.ndarray, np.ndarray)
        """
        ncol, nrow = disp['disparity_map'].shape
        offset = int((disp.attrs['window_size'] - 1) / 2)

        # Initialize ranges on max and min disparity values
        disp_max_range = np.ones((ncol, nrow), dtype=np.float32) * disp_max
        disp_min_range = np.ones((ncol, nrow), dtype=np.float32) * disp_min

        # Set invalid disparities as nan and store its indices
        tmp_disp_map = self.mask_invalid_disparities(disp)
        invalid_ind = np.where(np.isnan(tmp_disp_map))

        # Disparity windows
        disparity_windows = sliding_window(tmp_disp_map, (disp.attrs['window_size'], disp.attrs['window_size']))
        # Ignore warning in case the window is All-NaN
        warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')

        # To reduce memory, the data array is split (along the row axis) into multiple sub-arrays with a step of 100
        chunk_size = 100
        disp_chunked_y = np.array_split(disparity_windows, np.arange(chunk_size, ncol, chunk_size), axis=0)
        y_begin = offset

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
            # numpy.nanmin/nanmax : Compute the median along the specified axis, while ignoring NaNs
            for col in np.arange(len(disp_chunked_y)):
                # To reduce memory, the data array is split (along the col axis) into multiple sub-arrays,
                # with a step of 100
                disp_chunked_x = np.array_split(disp_chunked_y[col], np.arange(chunk_size, nrow, chunk_size), axis=1)
                x_begin = offset

                for row in np.arange(len(disp_chunked_x)):
                    y_end = y_begin + disp_chunked_y[col].shape[0]
                    x_end = x_begin + disp_chunked_x[row].shape[1]

                    disp_min_range[y_begin:y_end, x_begin:x_end] = np.nanmin(disp_chunked_x[row],
                                                                             axis=(2, 3)) - self._marge
                    disp_max_range[y_begin:y_end, x_begin:x_end] = np.nanmax(disp_chunked_x[row],
                                                                             axis=(2, 3)) + self._marge
                    x_begin += disp_chunked_x[row].shape[1]

                y_begin += disp_chunked_y[col].shape[0]

        del disparity_windows

        # Indices where disparity was invalid are set the max/min absolute value
        disp_min_range[invalid_ind] = disp_min
        disp_max_range[invalid_ind] = disp_max

        if self._scale_factor == 1:
            return disp_min_range, disp_max_range

        # Upsampling disparity range maps for next pyramid level
        disp_min_range = zoom(disp_min_range, self._scale_factor, order=0)
        disp_max_range = zoom(disp_max_range, self._scale_factor, order=0)

        return disp_min_range, disp_max_range



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to census method used in the cost volume measure step.
"""

from typing import Dict, Union, Tuple, List

import numpy as np
import xarray as xr
from json_checker import Checker, And

from pandora.img_tools import shift_right_img, census_transform
from pandora.matching_cost import matching_cost


@matching_cost.AbstractMatchingCost.register_subclass('census')
class Census(matching_cost.AbstractMatchingCost):
    """
    Census class allows to compute the cost volume
    """
    # Default configuration, do not change these values
    _WINDOW_SIZE = 5
    _SUBPIX = 1

    def __init__(self, **cfg: Dict[str, Union[str, int]]) -> None:
        """
        :param cfg: optional configuration,  {'window_size': value, 'subpix': value}
        :type cfg: dict
        :return: None
        """
        self.cfg = self.check_conf(**cfg)
        self._window_size = self.cfg['window_size']
        self._subpix = self.cfg['subpix']

    def check_conf(self, **cfg: Dict[str, Union[str, int]]) -> Dict[str, Union[str, int]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: matching cost configuration
        :type cfg: dict
        :return cfg: matching cost configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the conf
        if 'window_size' not in cfg:
            cfg['window_size'] = self._WINDOW_SIZE
        if 'subpix' not in cfg:
            cfg['subpix'] = self._SUBPIX

        schema = {
            'matching_cost_method': And(str, lambda input: 'census'),
            'window_size': And(int, lambda input: input in (3, 5)),
            'subpix': And(int, lambda input: input in (1, 2, 4))
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the matching cost method
        :return: None
        """
        print('census similarity measure')

    def compute_cost_volume(self, img_left: xr.Dataset, img_right: xr.Dataset, disp_min: int,
                            disp_max: int) -> xr.Dataset:
        """
        Computes the cost volume for a pair of images

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param disp_min: minimum disparity
        :type disp_min: int
        :param disp_max: maximum disparity
        :type disp_max: int
        :return: the cost volume dataset , with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :rtype: xarray.Dataset
        """
        # Contains the shifted right images
        img_right_shift = shift_right_img(img_right, self._subpix)

        # Maximal cost of the cost volume with census measure
        cmax = int(self._window_size ** 2)
        offset_row_col = int((self._window_size - 1) / 2)
        metadata = {'measure': 'census', 'subpixel': self._subpix,
                    'offset_row_col': offset_row_col, 'window_size': self._window_size,
                    'type_measure': 'min', 'cmax': cmax}

        # Apply census transformation
        left = census_transform(img_left, self._window_size)
        for i, img in enumerate(img_right_shift):
            img_right_shift[i] = census_transform(img, self._window_size)

        # Disparity range # pylint: disable=undefined-variable
        if self._subpix == 1:
            disparity_range = np.arange(disp_min, disp_max + 1)
        else:
            disparity_range = np.arange(disp_min, disp_max, step=1 / float(self._subpix))
            disparity_range = np.append(disparity_range, [disp_max])

        # Allocate the numpy cost volume cv = (disp, col, row), for efficient memory management
        cv = np.zeros((len(disparity_range), img_left['im'].shape[1], img_left['im'].shape[0]), dtype=np.float32)
        cv += np.nan

        # If offset, do not consider border position for cost computation
        if offset_row_col != 0:
            cv_crop= cv[:, offset_row_col:-offset_row_col, offset_row_col: -offset_row_col]
        else:
            cv_crop = cv

        # Giving the 2 images, the matching cost will be calculated as :
        #                 1, 1, 1                2, 5, 6
        #                 2, 1, 4                1, 1, 3
        #                 1, 3, 1                6, 1, 2
        #
        # for disp = 0, the costvolume will be
        #                        (1 - 2) (1 - 5) (1 - 6)
        #                        (2 - 1) (1 - 1) (4 - 3)
        #                        (1 - 6) (3 - 1) (1 - 2)
        # , we make the difference between the two full images
        #
        # for disp = -1, compute the difference between
        #                    1, 1                   2, 5
        #                    1, 4                   1, 1
        #                    3, 1                   6, 1
        # , the costvolume will be
        #                        nan  (1-2) (1-5)
        #                        nan  (1-1) (4-1)
        #                        nan  (3-6) (1-1)
        # , nan correspond to the first column of the left image that cannot be calculated,
        #  because there is no corresponding column for the disparity -1.
        #
        # In the following loop, the tuples point_p,point_q describe
        # to which part of the image the difference will be applied,
        # for disp = 0, we compute the difference on the whole images so point_p=(:,0:3) et point_q=(:,0:3),
        # for disp = -1, we use only a part of the images so point_p=(:,1:3) et point_q=(:,0:2)

        # Computes the matching cost
        # In the loop, cv is of shape (disp, col, row) and images / masks of shape (row, col)
        # np.swapaxes allow to interchange row and col in images and masks
        for disp in disparity_range:
            i_right = int((disp % 1) * self._subpix)
            point_p, point_q = self.point_interval(left, img_right_shift[i_right], disp)
            dsp = int((disp - disp_min) * self._subpix)

            cv_crop[dsp, point_p[0]:point_p[1], :] = np.swapaxes(
                self.census_cost(point_p, point_q, left, img_right_shift[i_right]), 0, 1)

        # Create the xarray.DataSet that will contain the cv of dimensions (row, col, disp)
        cv = self.allocate_costvolume(img_left, self._subpix, disp_min, disp_max, self._window_size, metadata,
                                      np.swapaxes(cv, 0, 2))

        # Remove temporary values
        del left, img_right_shift

        return cv

    def census_cost(self, point_p: Tuple[int, int], point_q: Tuple[int, int], img_left: xr.Dataset,
                    img_right: xr.Dataset) -> \
            List[int]:
        """
        Computes xor pixel-wise between pre-processed images by census transform

        :param point_p: Point interval, in the left image, over which the squared difference will be applied
        :type point_p: tuple
        :param point_q: Point interval, in the right image, over which the squared difference will be applied
        :type point_q: tuple
        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :return: the xor pixel-wise between elements in the interval
        :rtype: numpy array
        """
        xor_ = img_left['im'].data[:, point_p[0]:point_p[1]].astype('uint32') ^ img_right['im'].data[:,
                                                                                point_q[0]:point_q[1]].astype('uint32')
        return list(map(self.popcount32b, xor_))

    @staticmethod
    def popcount32b(row: int) -> int:
        """
        Computes the Hamming weight for the input row,
        Hamming weight is the number of symbols that are different from the zero

        :param row: 32-bit integer
        :type row: int
        :return: the number of symbols that are different from the zero
        :rtype: int
        """
        row -= (row >> 1) & 0x55555555
        row = (row & 0x33333333) + ((row >> 2) & 0x33333333)
        row = (row + (row >> 4)) & 0x0f0f0f0f
        row += row >> 8
        row += row >> 16
        return row & 0x7f



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the cost volume measure step.
"""

import logging
from abc import ABCMeta, abstractmethod
from math import ceil, floor
from typing import Tuple, List, Union

import numpy as np
import xarray as xr
from scipy.ndimage.morphology import binary_dilation

from pandora.img_tools import compute_std_raster
from pandora.img_tools import shift_right_img


class AbstractMatchingCost():
    """
    Abstract Matching Cost class
    """
    __metaclass__ = ABCMeta

    matching_cost_methods_avail = {}

    def __new__(cls, **cfg: Union[str, int]) -> object:
        """
        Return the plugin associated with the matching_cost_method given in the configuration

        :param cfg: configuration {'matching_cost_method': value}
        :type cfg: dictionary
        """

        if cls is AbstractMatchingCost:
            if isinstance(cfg['matching_cost_method'], str):
                try:
                    return super(AbstractMatchingCost, cls).__new__(cls.matching_cost_methods_avail
                                                                    [cfg['matching_cost_method']])
                except KeyError:
                    logging.error('No matching_cost method named % supported', cfg['matching_cost_method'])
                    raise KeyError
            else:
                if isinstance(cfg['matching_cost_method'], unicode):  # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractMatchingCost, cls).__new__(
                            cls.matching_cost_methods_avail[cfg['matching_cost_method'].encode('utf-8')])
                    except KeyError:
                        logging.error('No matching_cost method named % supported', cfg['matching_cost_method'])
                        raise KeyError
        else:
            return super(AbstractMatchingCost, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str, *args):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        :param args: allows to register one plugin that contains different methods
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.matching_cost_methods_avail[short_name] = subclass
            for arg in args:
                cls.matching_cost_methods_avail[arg] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self) -> None:
        """
        Describes the matching cost method
        :return: None
        """
        print('Matching cost description')

    @abstractmethod
    def compute_cost_volume(self, img_left: xr.Dataset, img_right: xr.Dataset, disp_min: int, disp_max: int
                            ) -> xr.Dataset:
        """
        Computes the cost volume for a pair of images

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset  containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param disp_min: minimum disparity
        :type disp_min: int
        :param disp_max: maximum disparity
        :type disp_max: int
        :return: the cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :rtype: xarray.Dataset
        """

    @staticmethod
    def allocate_costvolume(img_left: xr.Dataset, subpix: int, disp_min: int, disp_max: int, window_size: int,
                            metadata: dict, np_data: np.ndarray = None) -> xr.Dataset:
        """
        Allocate the cost volume

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param subpix: subpixel precision = (1 or 2 or 4)
        :type subpix: int
        :param disp_min: minimum disparity
        :type disp_min: int
        :param disp_max: maximum disparity
        :type disp_max: int
        :param window_size: size of the aggregation window
        :type window_size: int, odd number
        :param metadata: dictionary storing arbitrary metadata
        :type metadata: dictionary
        :param np_data: the array’s data
        :type np_data: 3D numpy array, dtype=np.float32
        :return: the dataset cost volume with the cost_volume and the confidence measure with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :rtype: xarray.Dataset
        """
        c_row = img_left['im'].coords['row']
        c_col = img_left['im'].coords['col']

        # First pixel in the image that is fully computable (aggregation windows are complete)
        offset_row_col = int((window_size - 1) / 2)
        row = np.arange(c_row[0], c_row[-1] + 1)
        col = np.arange(c_col[0], c_col[-1] + 1)
        row_off = np.arange(c_row[0] + offset_row_col, c_row[-1] - offset_row_col + 1)
        col_off = np.arange(c_col[0] + offset_row_col, c_col[-1] - offset_row_col + 1)

        # Compute the disparity range
        if subpix == 1:
            disparity_range = np.arange(disp_min, disp_max + 1)
        else:
            disparity_range = np.arange(disp_min, disp_max, step=1 / float(subpix))
            disparity_range = np.append(disparity_range, [disp_max])

        # Create the cost volume
        if np_data is None:
            np_data = np.zeros((len(row), len(col), len(disparity_range)), dtype=np.float32)

        cost_volume = xr.Dataset({'cost_volume': (['row', 'col', 'disp'], np_data)},
                                 coords={'row': row, 'col': col, 'disp': disparity_range})
        cost_volume.attrs = metadata

        # Allocate the confidence measure in the cost volume Dataset
        confidence_measure = compute_std_raster(img_left, window_size).reshape((len(row_off), len(col_off), 1))
        # Create the confidence measure with the original image size and fill it
        confidence_measure_full = np.full((len(row), len(col), 1), np.nan, dtype=np.float32)
        if offset_row_col != 0:
            confidence_measure_full[offset_row_col: - offset_row_col, offset_row_col: - offset_row_col, :] = \
                confidence_measure
        else:
            confidence_measure_full = confidence_measure

        cost_volume = cost_volume.assign_coords(indicator=['stereo_pandora_intensityStd'])
        cost_volume['confidence_measure'] = xr.DataArray(data=confidence_measure_full.astype(np.float32),
                                                         dims=['row', 'col', 'indicator'])

        return cost_volume

    @staticmethod
    def point_interval(img_left: xr.Dataset, img_right: xr.Dataset, disp: float) -> \
            Tuple[Tuple[int, int], Tuple[int, int]]:
        """
        Computes the range of points over which the similarity measure will be applied

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param disp: current disparity
        :type disp: float
        :return: the range of the left and right image over which the similarity measure will be applied
        :rtype: tuple
        """
        _, nx_left = img_left['im'].shape
        _, nx_right = img_right['im'].shape

        # range in the left image
        point_p = (max(0 - disp, 0), min(nx_left - disp, nx_left))
        # range in the right image
        point_q = (max(0 + disp, 0), min(nx_right + disp, nx_right))

        # Because the disparity can be floating
        if disp < 0:
            point_p = (int(ceil(point_p[0])), int(ceil(point_p[1])))
            point_q = (int(ceil(point_q[0])), int(ceil(point_q[1])))
        else:
            point_p = (int(floor(point_p[0])), int(floor(point_p[1])))
            point_q = (int(floor(point_q[0])), int(floor(point_q[1])))

        return point_p, point_q

    @staticmethod
    def masks_dilatation(img_left: xr.Dataset, img_right: xr.Dataset, window_size: int,
                         subp: int) -> Tuple[xr.DataArray, List[xr.DataArray]]:
        """
        Return the left and right mask with the convention :
            - Invalid pixels are nan
            - No_data pixels are nan
            - Valid pixels are 0

        Apply dilation on no_data : if a pixel contains a no_data in its aggregation window, then the central pixel
        becomes no_data

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param window_size: window size of the measure
        :type window_size: int
        :param subp: subpixel precision = (1 or 2 or 4)
        :type subp: int
        :return: the left mask and the right masks:

                - left mask :  xarray.DataArray msk 2D(row, col)
                - right mask :  xarray.DataArray msk 2D(row, col)
                - right mask shifted :  xarray.DataArray msk 2D(row, shifted col by 0.5)
        :rtype: tuple (left mask, list[right mask, right mask shifted by 0.5])
        """
        # Create the left mask with the convention : 0 = valid, nan = invalid and no_data
        if 'msk' in img_left.data_vars:
            dilatate_left_mask = np.zeros(img_left['msk'].shape)
            # Invalid pixels are nan
            dilatate_left_mask[
                np.where((img_left['msk'].data != img_left.attrs['valid_pixels']) & (
                        img_left['msk'].data != img_left.attrs['no_data_mask']))] = np.nan
            # Dilatation : pixels that contains no_data in their aggregation window become no_data = nan
            dil = binary_dilation(img_left['msk'].data == img_left.attrs['no_data_mask'],
                                  structure=np.ones((window_size, window_size)), iterations=1)
            dilatate_left_mask[dil] = np.nan
        else:
            # All pixels are valid
            dilatate_left_mask = np.zeros(img_left['im'].shape)

        # Create the right mask with the convention : 0 = valid, nan = invalid and no_data
        if 'msk' in img_right.data_vars:
            dilatate_right_mask = np.zeros(img_right['msk'].shape)
            # Invalid pixels are nan
            dilatate_right_mask[
                np.where((img_right['msk'].data != img_right.attrs['valid_pixels']) & (
                        img_right['msk'].data != img_right.attrs['no_data_mask']))] = np.nan
            # Dilatation : pixels that contains no_data in their aggregation window become no_data = nan
            dil = binary_dilation(img_right['msk'].data == img_right.attrs['no_data_mask'],
                                  structure=np.ones((window_size, window_size)), iterations=1)
            dilatate_right_mask[dil] = np.nan

        else:
            # All pixels are valid
            dilatate_right_mask = np.zeros(img_left['im'].shape)

        ny_, nx_ = img_left['im'].shape
        row = np.arange(0, ny_)
        col = np.arange(0, nx_)

        # Shift the right mask, for sub-pixel precision. If an no_data or invalid pixel was used to create the
        # sub-pixel point, then the sub-pixel point is invalidated / no_data.
        dilatate_right_mask_shift = xr.DataArray()
        if subp != 1:
            # Since the interpolation of the right image is of order 1, the shifted right mask corresponds
            # to an aggregation of two columns of the dilated right mask
            str_row, str_col = dilatate_right_mask.strides
            shape_windows = (dilatate_right_mask.shape[0], dilatate_right_mask.shape[1] - 1, 2)
            strides_windows = (str_row, str_col, str_col)
            aggregation_window = np.lib.stride_tricks.as_strided(dilatate_right_mask, shape_windows, strides_windows)
            dilatate_right_mask_shift = np.sum(aggregation_window, 2)

            # Whatever the sub-pixel precision, only one sub-pixel mask is created,
            # since 0.5 shifted mask == 0.25 shifted mask
            col_shift = np.arange(0 + 0.5, nx_ - 1, step=1)
            dilatate_right_mask_shift = xr.DataArray(dilatate_right_mask_shift,
                                                     coords=[row, col_shift], dims=['row', 'col'])

        dilatate_left_mask = xr.DataArray(dilatate_left_mask,
                                          coords=[row, col], dims=['row', 'col'])
        dilatate_right_mask = xr.DataArray(dilatate_right_mask,
                                           coords=[row, col], dims=['row', 'col'])

        return dilatate_left_mask, [dilatate_right_mask, dilatate_right_mask_shift]

    @staticmethod
    def dmin_dmax(disp_min: Union[int, np.ndarray], disp_max: Union[int, np.ndarray]) -> Tuple[int, int]:
        """
        Find the smallest disparity present in disp_min, and the highest disparity present in disp_max

        :param disp_min: minimum disparity
        :type disp_min: int or np.ndarray
        :param disp_max: maximum disparity
        :type disp_max: int or np.ndarray
        :return: dmin_min: the smallest disparity in disp_min, dmax_max: the highest disparity in disp_max
        :rtype: Tuple(int, int)
        """
        # Disp_min is a fixed disparity
        if isinstance(disp_min, int):
            dmin_min = disp_min

        # Disp_min is a variable disparity
        else:
            dmin_min = int(np.nanmin(disp_min))

        # Disp_max is a fixed disparity
        if isinstance(disp_max, int):
            dmax_max = disp_max

        # Disp_max is a variable disparity
        else:
            dmax_max = int(np.nanmax(disp_max))

        return dmin_min, dmax_max

    def cv_masked(self, img_left: xr.Dataset, img_right: xr.Dataset, cost_volume: xr.Dataset,
                  disp_min: Union[int, np.ndarray], disp_max: Union[int, np.ndarray]) -> None:
        """
        Masks the cost volume :
            - costs which are not inside their disparity range, are masked with a nan value
            - costs of invalid_pixels (invalidated by the input image mask), are masked with a nan value
            - costs of no_data pixels, are masked with a nan value. If a valid pixel contains a no_data in its
                aggregation window, then the cost of the central pixel is masked with a nan value

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cost_volume: the cost_volume DataSet with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cost_volume: xarray.Dataset
        :param disp_min: minimum disparity
        :type disp_min: int or np.ndarray
        :param disp_max: maximum disparity
        :type disp_max: int or np.ndarray
        :param cfg: images configuration containing the mask convention : valid_pixels, no_data
        :type cfg: dict
        :return: None
        """
        ny_, nx_, nd_ = cost_volume['cost_volume'].shape # pylint: disable=unused-variable

        dmin, dmax = self.dmin_dmax(disp_min, disp_max) # pylint: disable=unused-variable

        # ----- Masking invalid pixels -----

        # Contains the shifted right images
        img_right_shift = shift_right_img(img_right, self._subpix)

        # Computes the validity mask of the cost volume : invalid pixels or no_data are masked with the value nan.
        # Valid pixels are = 0
        mask_left, mask_right = self.masks_dilatation(img_left, img_right, self._window_size, self._subpix)

        for disp in cost_volume.coords['disp'].data:
            i_right = int((disp % 1) * self._subpix)
            point_p, point_q = self.point_interval(img_left, img_right_shift[i_right], disp)

            # Point interval in the left image
            p_mask = (point_p[0], point_p[1])
            # Point interval in the right image
            q_mask = (point_q[0], point_q[1])

            i_mask_right = min(1, i_right)
            dsp = int((disp - dmin) * self._subpix)

            # Invalid costs in the cost volume
            cost_volume['cost_volume'].data[:, p_mask[0]:p_mask[1], dsp] += (
                    mask_right[i_mask_right].data[:, q_mask[0]:q_mask[1]] + mask_left.data[:, p_mask[0]:p_mask[1]])

        # ----- Masking disparity range -----

        # Fixed range of disparities
        if isinstance(disp_min, int) and isinstance(disp_max, int):
            return

        # Disparity range may be one size bigger in y axis
        if disp_min.shape[0] > ny_:
            disp_min = disp_min[0: ny_, :]
            disp_max = disp_max[0: ny_, :]
        if disp_min.shape[1] > nx_:
            disp_min = disp_min[:, 0: nx_]
            disp_max = disp_max[:, 0: nx_]

        # Mask the costs computed with a disparity lower than disp_min and higher than disp_max
        for dsp in range(nd_):
            masking = np.where(np.logical_or(cost_volume.coords['disp'].data[dsp] < disp_min,
                                             cost_volume.coords['disp'].data[dsp] > disp_max))
            cost_volume['cost_volume'].data[masking[0], masking[1], dsp] = np.nan



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the vfit method used in the refinement step.
"""

from typing import Dict, Tuple

import numpy as np
from json_checker import Checker, And
from numba import njit

import pandora.constants as cst
from . import refinement


@refinement.AbstractRefinement.register_subclass('vfit')
class Vfit(refinement.AbstractRefinement):
    """
    Vfit class allows to perform the subpixel cost refinement step
    """

    def __init__(self, **cfg: str) -> None:
        """
        :param cfg: optional configuration, {}
        :type cfg: dict
        :return: None
        """
        self.cfg = self.check_conf(**cfg)
        self._refinement_method_name = str(self.cfg['refinement_method'])

    @staticmethod
    def check_conf(**cfg: str) -> Dict[str, str]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: refinement configuration
        :type cfg: dict
        :return cfg: refinement configuration updated
        :rtype: dict
        """
        schema = {
            'refinement_method': And(str, lambda x: 'vfit')
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the subpixel refinement method
        :return: None
        """
        print('Vfit refinement method')

    @staticmethod
    @njit()
    def refinement_method(cost: np.ndarray, disp: float, measure: str) -> Tuple[float, float, int]:
        """
        Return the subpixel disparity and cost, by matching a symmetric V shape (linear interpolation)

        :param cost: cost of the values disp - 1, disp, disp + 1
        :type cost: 1D numpy array : [cost[disp -1], cost[disp], cost[disp + 1]]
        :param disp: the disparity
        :type disp: float
        :param measure: the type of measure used to create the cost volume
        :param measure: string = min | max
        :return: the refined disparity (disp + sub_disp), the refined cost and the state of the pixel( Information: \
        calculations stopped at the pixel step, sub-pixel interpolation did not succeed )
        :rtype: float, float, int
        """
        if (np.isnan(cost[0])) or (np.isnan(cost[2])):
            # Information: calculations stopped at the pixel step, sub-pixel interpolation did not succeed
            return disp, cost[1], cst.PANDORA_MSK_PIXEL_STOPPED_INTERPOLATION

        inverse = 1
        if measure == 'max':
            # Additive inverse : if a < b then -a > -b
            inverse = -1

        # The problem is to approximate sub_cost function with an affine function: y = a * x + origin
        # Calculate the slope
        a = cost[2] - cost[1]

        # Compare the difference disparity between (cost[0]-cost[1]) and (cost[2]-cost[1]): the highest cost is used
        if (inverse * cost[0]) > (inverse * cost[2]):
            a = cost[0] - cost[1]

        if abs(a) < 1.e-15:
            return disp, cost[1], 0

        # Problem is resolved with tangents equality, due to the symmetric V shape of 3 points (cv0, cv2 and (x,y))
        # sub_disp is dx
        sub_disp = ((cost[0] - cost[2]) / (2 * a))

        # sub_cost is y
        sub_cost = a * (sub_disp - 1) + cost[2]

        return sub_disp + disp, sub_cost, 0



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains classes and functions associated to the validation step.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Dict, Union

import numpy as np
import xarray as xr
from json_checker import Checker, And, Or, OptionalKey

import pandora.constants as cst
from pandora import common


class AbstractValidation():
    """
    Abstract Validation class
    """
    __metaclass__ = ABCMeta

    validation_methods_avail = {}

    def __new__(cls, **cfg: dict) -> object:
        """
        Return the plugin associated with the validation_method given in the configuration

        :param cfg: configuration {'validation_method': value}
        :type cfg: dictionary
        """
        if cls is AbstractValidation:
            if isinstance(cfg['validation_method'], str):
                try:
                    return super(AbstractValidation, cls).__new__(
                        cls.validation_methods_avail[cfg['validation_method']])
                except KeyError:
                    logging.error('No validation method named % supported', cfg['validation_method'])
                    raise KeyError
            else:
                if isinstance(cfg['validation_method'], unicode): # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractValidation, cls).__new__(
                            cls.validation_methods_avail[cfg['validation_method'].encode('utf-8')])
                    except KeyError:
                        logging.error('No validation method named % supported', cfg['validation_method'])
                        raise KeyError
        else:
            return super(AbstractValidation, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.validation_methods_avail[short_name] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self) -> None:
        """
        Describes the validation method
        :return: None
        """
        print('Validation method description')

    @abstractmethod
    def disparity_checking(self, dataset_left: xr.Dataset, dataset_right: xr.Dataset, img_left: xr.Dataset = None,
                           img_right: xr.Dataset = None, cv: xr.Dataset = None) -> xr.Dataset:
        """
        Determination of occlusions and false matches by performing a consistency check on valid pixels.
        Update the validity_mask :

            - If out & MSK_PIXEL_OCCLUSION != 0 : Invalid pixel : occluded pixel
            - If out & MSK_PIXEL_MISMATCH  != 0  : Invalid pixel : mismatched pixel

        | Update the measure map: add the disp RL / disp LR distances

        :param dataset_left: left Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type dataset_left: xarray.Dataset
        :param dataset_right: right Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type dataset_right: xarray.Dataset
        :param img_left: left Datset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost_volume Dataset with the variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: the left dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col) with the bit 8 and 9 of the validity_mask :
                - If out & MSK_PIXEL_OCCLUSION != 0 : Invalid pixel : occluded pixel
                - If out & MSK_PIXEL_MISMATCH  != 0  : Invalid pixel : mismatched pixel
        :rtype: xarray.Dataset
        """


@AbstractValidation.register_subclass('cross_checking')
class CrossChecking(AbstractValidation):
    """
    CrossChecking class allows to perform the validation step
    """
    # Default configuration, do not change this value
    _THRESHOLD = 1.

    def __init__(self, **cfg) -> None:
        """
        :param cfg: optional configuration, {'cross_checking_threshold': value,
                                            'interpolated_disparity': value, 'filter_interpolated_disparities': value}
        :type cfg: dictionary
        :return: None
        """
        self.cfg = self.check_conf(**cfg)
        self._threshold = self.cfg['cross_checking_threshold']

    def check_conf(self, **cfg: Union[str, int, float, bool]) -> Dict[str, Union[str, int, float, bool]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: optimization configuration
        :type cfg: dict
        :return: optimization configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'cross_checking_threshold' not in cfg:
            cfg['cross_checking_threshold'] = self._THRESHOLD
        if 'right_left_mode' not in cfg:
            cfg['right_left_mode'] = 'accurate'

        schema = {
            'validation_method': And(str, lambda input: 'cross_checking'),
            'cross_checking_threshold': Or(int, float),
            'right_left_mode': And(str, lambda input: common.is_method(input, ['accurate', 'approximate'])),
            OptionalKey('interpolated_disparity'): And(str, lambda input: common.is_method(input, ['mc-cnn', 'sgm']))
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the validation method
        :return: None
        """
        print('Cross-checking method')

    def disparity_checking(self, dataset_left: xr.Dataset, dataset_right: xr.Dataset, img_left: xr.Dataset = None,
                           img_right: xr.Dataset = None, cv: xr.Dataset = None) -> xr.Dataset:
        """
        Determination of occlusions and false matches by performing a consistency check on valid pixels.

        Update the validity_mask :

            - If out & MSK_PIXEL_OCCLUSION != 0 : Invalid pixel : occluded pixel
            - If out & MSK_PIXEL_MISMATCH  != 0  : Invalid pixel : mismatched pixel

        | Update the measure map: add the disp RL / disp LR distances

        :param dataset_left: left Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type dataset_left: xarray.Dataset
        :param dataset_right: right Dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col)
        :type dataset_right: xarray.Dataset
        :param img_left: left Datset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost_volume Dataset with the variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: the left dataset with the variables :

            - disparity_map 2D xarray.DataArray (row, col)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
            - validity_mask 2D xarray.DataArray (row, col) with the bit 8 and 9 of the validity_mask :

                - If out & MSK_PIXEL_OCCLUSION != 0 : Invalid pixel : occluded pixel
                - If out & MSK_PIXEL_MISMATCH  != 0  : Invalid pixel : mismatched pixel
        :rtype: xarray.Dataset
        """
        nb_row, nb_col, nb_indicator = dataset_left['confidence_measure'].shape
        disparity_range = np.arange(dataset_left.attrs['disp_min'], dataset_left.attrs['disp_max'] + 1)

        # Add a new indicator to the confidence measure DataArray
        conf_measure = np.full((nb_row, nb_col, nb_indicator + 1), np.nan, dtype=np.float32)
        conf_measure[:, :, :-1] = dataset_left['confidence_measure'].data

        indicator = np.copy(dataset_left.coords['indicator'])
        indicator = np.append(indicator, 'validation_pandora_distanceOfDisp')

        # Remove confidence_measure dataArray from the dataset to update it
        dataset_left = dataset_left.drop_dims('indicator')
        dataset_left = dataset_left.assign_coords(indicator=indicator)
        dataset_left['confidence_measure'] = xr.DataArray(data=conf_measure, dims=['row', 'col', 'indicator'])

        for row in range(0, nb_row):
            # Exclude invalid pixel :
            valid_pixel = np.where((dataset_left['validity_mask'].data[row, :] & cst.PANDORA_MSK_PIXEL_INVALID) == 0)

            col_left = np.arange(nb_col, dtype=np.int)
            col_left = col_left[valid_pixel]

            col_right = col_left + dataset_left['disparity_map'].data[row, col_left]
            # Round elements of the array to the nearest integer
            col_right = np.rint(col_right).astype(int)

            # Left-Right consistency, for pixel i :
            # If | Disp_right(i + rint(Disp_left(i)) + Disp_left(i) | > self._threshold :
            # i is invalid, mismatched or occlusion detected
            # If | Disp_right(i + rint(Disp_left(i)) + Disp_left(i) | <= self._threshold : i is valid

            # Apply cross checking on pixels i + round(Disp_left(i) inside the right image
            inside_right = np.where((col_right >= 0) & (col_right < nb_col))

            # Conversion from nan to inf
            right_disp = dataset_right['disparity_map'].data[row, col_right[inside_right]]
            right_disp[np.isnan(right_disp)] = np.inf
            left_disp = dataset_left['disparity_map'].data[row, col_left[inside_right]]
            left_disp[np.isnan(left_disp)] = np.inf

            # Allocate to the measure map, the distance disp LR / disp RL indicator
            dataset_left['confidence_measure'].data[row, col_left[inside_right], -1] = np.abs(right_disp + left_disp)

            # left image pixels invalidated by the cross checking
            invalid = np.abs(right_disp + left_disp) > self._threshold

            # Detect mismatched and occlusion :
            # For a left image pixel i invalidated by the cross checking :
            # mismatch if : Disp_right(i + d) = -d, for any other d
            # occlusion otherwise

            # Index : i + d, for any other d. 2D np array (nb invalid pixels, nb disparity )
            index = np.tile(disparity_range, (len(col_left[inside_right][invalid]), 1)).astype(np.float32) + \
                    np.tile(col_left[inside_right][invalid], (len(disparity_range), 1)).transpose()

            inside_col_disp = np.where((index >= 0) & (index < nb_col))

            # disp_right : Disp_right(i + d)
            disp_right = np.full(index.shape, np.inf, dtype=np.float32)
            disp_right[inside_col_disp] = dataset_right['disparity_map'].data[row, index[inside_col_disp].astype(int)]

            # Check if rint(Disp_right(i + d)) == -d
            comp = (np.rint(disp_right) == np.tile(-1 * disparity_range,
                                                   (len(col_left[inside_right][invalid]), 1)).astype(
                np.float32))
            comp = np.sum(comp, axis=1)
            comp[comp > 1] = 1

            dataset_left['validity_mask'].data[row, col_left[inside_right][invalid]] += cst.PANDORA_MSK_PIXEL_OCCLUSION
            dataset_left['validity_mask'].data[row, col_left[inside_right][invalid]] += \
                (cst.PANDORA_MSK_PIXEL_MISMATCH * comp).astype(np.uint16)
            dataset_left['validity_mask'].data[row, col_left[inside_right][invalid]] -= \
                (cst.PANDORA_MSK_PIXEL_OCCLUSION * comp).astype(np.uint16)

            # Pixels i + round(Disp_left(i) outside the right image are occlusions
            outside_right = np.where((col_right < 0) & (col_right >= nb_col))
            dataset_left['validity_mask'].data[row, col_left[outside_right]] += cst.PANDORA_MSK_PIXEL_OCCLUSION

        dataset_left.attrs['validation'] = 'cross_checking'

        return dataset_left
